%{
/**
 | Lexical analyzer for desktop calculator
 | From the bison manual example, modified by MLO
 |
 | $Id: dtc.lex,v 2.7 2002/01/19 16:25:02 loreti Exp $
**/

#ifdef USE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif /* USE_READLINE */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dtcdef.h"
#include "dtc.h"

extern int     interactive;
extern symRec *root;
extern char   *prompt;
extern char   *local;
extern size_t  length;

static int     getInput(char *, const size_t);
static int     normalInput(char *, const size_t);
static char   *stripString(char *, size_t *);

#define YY_INPUT(b,r,l)         { r = getInput(b, l); }
%}

SGN [+-]
DGT [0-9]
LTR [a-zA-Z]
HEX [0-9a-fA-F]
OCT [0-7]
SYM [a-zA-Z0-9_]
EXP [eE]{SGN}?{DGT}+

%%

<<EOF>> { return 0; }

[\t ] { }

{LTR}{SYM}* {
  symRec *p;
  int newNode;

  root = insertNode(root, yytext, &newNode, &p);
  if (newNode) {
    p->type      = SYMBOL;
    p->value.var = 0.0;
  }
  yylval.tptr = p;
  return p->type;
}

"0" {
  yylval.val = 0;
  return NUMBER;
}

(0[xX]{HEX}+)|(0{OCT}+)|([1-9]{DGT}*) {
  yylval.val = strtoul(yytext, 0, 0);
  return NUMBER;
}

({DGT}+{EXP})|(({DGT}+\.)|({DGT}*\.{DGT}+)){EXP}? {
  yylval.val = strtod(yytext, 0);
  return NUMBER;
}

"\n"|";"|"+"|"-"|"*"|"/"|"%"|"^"|"="|"("|")" {
  return yytext[0];
}

"?" {
  puts("  This is dtc version 9  -  MLO 19-Jan-2002");
  puts("  - Statement:\t\t{ expression | variable=expression | variable }");
  puts("  - Input syntax:\tstatement [ ; statement [ ; ... ]]");
  puts("  - Known functions:\tsin, cos, tan, asin, acos, atan, ln, log, exp"
       ", sqrt,");
  puts("\t\t\ttrunc, frac, round, abs");
  puts("  - Known constants:\tpi, e, dtor, rtod");
  puts("  - Floating input:\tdecimal (1.2) or scientific (3.4E5) notation");
  puts("  - Integer input:\tdecimal (67), hex (0xA) or octal (012) notation");
  puts("  - Special functions:\tlast, hex, oct, stab; "
       "? for help; EOF to exit");
}

. { int c;
 puts("  Error: syntax error");
 while ((c=input()) != EOF) {
   if (c == ';'   ||   c == '\n') return c;
 }
 return 0;
}

%%

#ifdef USE_READLINE
static int getInput(
  char         *buffer,
  const size_t  maxChar
) {

/**
 | Local function to get input through the GNU "readline" interface.
 | We are required to move at most "maxChar" characters from the input
 | stream to "buffer", and to return the number of characters moved or
 | YY_NULL if an end-of-file has been detected.
 |
 | Local variables:
 |
 | local  : local buffer where the input characters will be moved;
 | length : current size of the local buffer;
 | avail  : number of characters available to be moved;
 | next   : pointer to the next character to be moved;
 | result : value to be returned.
**/

  static size_t  avail  = 0;
  static char   *next;
  int            result;

  if (!interactive) return normalInput(buffer, maxChar);

  while (avail == 0) {

/**
 | If the local buffer is empty, asks for an input line; if the line
 | is not empty, is pushed to the history buffer and copied to the
 | local storage area with a trailing newline.
**/
    char   *line;
    char   *start;
    size_t  count;

    if ((line = readline(prompt)) == 0) return YY_NULL;

    if (*(start = stripString(line, &count))) {
      add_history(start);
      if (length < count) {
        if ((local = realloc(local, count+2)) == 0) {
          fputs("No heap memory available\n", stderr);
          exit(EXIT_FAILURE);
        }
        length = count;
      }
      next = memcpy(local, start, count);
      strcpy(local+count, "\n");
      avail = count+1;
    }
    free(line);
  }

  result = (avail > maxChar ? maxChar : avail);
  if (result) {
    memcpy(buffer, next, result);
    next  += result;
    avail -= result;
  }
  return result;
}

static char *stripString(
  char   *line,
  size_t *pCount
) {

/**
 | Returns a pointer to the first non-whitespace character of
 | "line"; stores in "*pCount" the number of characters up to
 | the last non-whitespace in "line".
**/

  register char c, *pc, *last;

  for (;;line++) {
    if ((c = *line) == '\0') {
      *pCount = 0;
      return line;
    }
    if (!isspace((unsigned int) c)) break;
  }

  for (pc = last = line+1;   (c = *pc++) != '\0';   ) {
    if (!isspace((unsigned int) c)) last = pc;
  }

  *pCount = last - line;
  return line;
}
#else  /* USE_READLINE */
static int getInput(
  char         *buffer,
  const size_t  maxChar
) {
  return normalInput(buffer, maxChar);
}
#endif /* USE_READLINE */

static int normalInput(
  char         *buffer,
  const size_t  maxChar
) {
  static size_t  avail = 0;
  static char   *next;
  int            result;

  if (avail == 0) {
    if (prompt) fputs(prompt, stdout);
    if ((next = fgets(local, length+2, stdin)) == 0) return YY_NULL;
    if (!interactive) fputs(local, stdout);
    avail = strlen(local);
  }

  result = (avail > maxChar ? maxChar : avail);
  if (result) {
    memcpy(buffer, next, result);
    next  += result;
    avail -= result;
  }
  return result;
}
