%{
/**
 | Desktop calculator grammar
 | From the bison manual example, modified by MLO
 |
 | $Id: dtc.y,v 2.5 2005/05/05 07:12:19 loreti Exp $
**/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef USE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif /* USE_READLINE */
#include "dtcdef.h"

#define YYDEBUG 0                     /* 1 for bison debug info on stderr  */

static double *last;                  /* Pointer to the last printed value */

static void hex(void);                /* dec -> hex/oct conversion         */
static void oct(void);

static void printTree(void);          /* Symbol table printout: tree,      */
static void pN(const symRec *);       /*   and single nodes                */

static double myAsin(const double);   /* Wrapper functions, that check the */
static double myAcos(const double);   /*   argument before calling the     */
static double myLog(const double);    /*     corresponding math library    */
static double myLog10(const double);  /*       procedure                   */
static double mySqrt(const double);

static double myTrunc(const double);  /* Integer part, fractional part and */
static double frac(const double);     /*   rounding to the nearer integer, */
static double vround(const double);   /*     absolute value                */
static double myAbs(const double);

/**
 | Global variables
**/

int     interactive;
symRec *root   = 0;
char   *prompt = 0;
char   *local;
size_t  length = 510;

%}

%union {
  double val;
  symRec *tptr;
};                              /* YYSTYPE, possible return values */

%token <val>  NUMBER            /* Basic tokens: numbers, symbols, */
%token <tptr> SYMBOL            /*   functions, expressions,       */
%token <tptr> FUNCTION          /*     special conversions.        */
%token <tptr> SPECIAL
%type  <val>  exp

%right '='                      /* Operators, left- and right- */
%left  '-' '+'                  /*   associative, from low to  */
%left  '*' '/' '%'              /*   high precedence order.    */
%right '^'

%%

input:  /* empty */ | input line ;

line:   '\n'                                                               ;
line:   ';'                                                                ;
line:   error '\n'                                                         ;
line:   error ';'                                                          ;
line:   exp '\n' { printf ("  %.10g\n", (*last=$1)); }                     ;
line:   exp ';'  { printf ("  %.10g\n", (*last=$1)); }                     ;
line:   SPECIAL '\n' { (*($1->value.v_funptr))(); }                        ;
line:   SPECIAL ';'  { (*($1->value.v_funptr))(); }                        ;

exp:    NUMBER                   { $$ = $1; }                              ;
exp:    '-' NUMBER               { $$ = -$2;}                              ;
exp:    SYMBOL '=' exp           { $1->value.var = $3;  $$ = $3; }         ;
exp:    SYMBOL                   { $$ = $1->value.var; }                   ;
exp:    '-' SYMBOL               { $$ = -$2->value.var; }                  ;
exp:    FUNCTION '(' exp ')'     { $$ = (*($1->value.d_funptr))($3); }     ;
exp:    '-' FUNCTION '(' exp ')' { $$ = -(*($2->value.d_funptr))($4); }    ;
exp:    FUNCTION                 { $$ = (*($1->value.d_funptr))(*last); }  ;
exp:    '-' FUNCTION             { $$ = -(*($2->value.d_funptr))(*last); } ;
exp:    exp '+' exp              { $$ = $1 + $3; }                         ;
exp:    exp '-' exp              { $$ = $1 - $3; }                         ;
exp:    exp '*' exp              { $$ = $1 * $3; }                         ;
exp:    exp '/' exp              { if ($3 == 0) {
                                     puts("  Error: division by zero");
                                     $$ = 0;
                                   } else {
                                     $$ = $1 / $3;
                                   }
                                 }                                         ;
exp:    exp '%' exp              { if ($3 == 0) {
                                     puts("  Error: division by zero");
                                     $$ = 0;
                                   } else {
                                     int i = $1 / $3;
                                     $$ = $1 - i * $3;
                                   }
                                 }                                         ;
exp:    exp '^' exp              { $$ = pow($1, $3); }                     ;
exp:    '(' exp ')'              { $$ = $2; }                              ;

%%

/* Main program */

static funTable funs[] = {              /* Predefined functions */
  {"sin",    sin},
  {"cos",    cos},
  {"tan",    tan},
  {"asin",   myAsin},
  {"acos",   myAcos},
  {"atan",   atan},
  {"ln",     myLog},
  {"log",    myLog10},
  {"exp",    exp},
  {"sqrt",   mySqrt},
  {"trunc",  myTrunc},
  {"round",  vround},
  {"frac",   frac},
  {"abs",    myAbs},
  {0, 0}
};

static constTable cons[] = {            /* Predefined constants */
  {"pi",     3.1415926535897932385},
  {"e",      2.7182818284590452354},
  {"dtor",   0.01745329252},
  {"rtod",   57.29577951},
  {"last",   0},                        /* _MUST_BE_ the last one */
  {0, 0}
};

static specTable spec[] = {             /* Special commands */
  {"hex",    hex},
  {"oct",    oct},
  {"stab",   printTree},
  {0, 0}
};

static void initTable(void);

int main()
{
  interactive = isatty(STDIN_FILENO);
#ifdef USE_READLINE
  if (interactive) rl_bind_key('\t', rl_insert);
#endif /* USE_READLINE */

  initTable();

#if YYDEBUG != 0
  yydebug = YYDEBUG;
#endif

  if ((local = malloc(length+2)) == 0) {
    fputs("No heap memory available\n", stderr);
    exit(EXIT_FAILURE);
  }
  if (interactive) prompt = "? ";
  yyparse();
  if (interactive) puts("");
  return EXIT_SUCCESS;
}

static void initTable(void)
{
  symRec *p;
  int     i;

  for (i = 0;   funs[i].name != 0;   ++i) {
    root              = insertNode(root, funs[i].name, 0, &p);
    p->type           = FUNCTION;
    p->value.d_funptr = funs[i].fun;
  }

/**
 | The value field for the last constant entered is pointed to by "last";
 | *last is updates after every printed expression, and the constant
 | "last" recalls its value.
**/

  for (i = 0;   cons[i].name != 0;   ++i) {
    root                      = insertNode(root, cons[i].name, 0, &p);
    p->type                   = SYMBOL;
    *(last = &(p->value.var)) = cons[i].value;
  }

  for (i = 0;   spec[i].name != 0;   ++i) {
    root              = insertNode(root, spec[i].name, 0, &p);
    p->type           = SPECIAL;
    p->value.v_funptr = spec[i].fun;
  }
}

symRec *insertNode(
        symRec  *pNode,
  const char    *name,
        int     *pNew,
        symRec **p
) {

/**
 | Looks recursively, in the binary tree containing the symbol table,
 | for a node containing the ASCII string "name"; "pNode" points to
 | the node we are looking at currently.  If an existing node cannot
 | be found, an empty new one is created and inserted in the tree; a
 | pointer to the target node will be stored in "*p".  If "pNew" is
 | not NULL, in *pNew will be stored or 0 (an existing node has been
 | found) or 1 (a new node has been created).
**/

  if (pNode) {
    int compare;

    if ((compare = strcmp(name, pNode->name)) == 0) {
      if (pNew) *pNew = 0;
      *p = pNode;
    } else if (compare < 0) {
      pNode->left = insertNode(pNode->left, name, pNew, p);
    } else {
      pNode->right = insertNode(pNode->right, name, pNew, p);
    }
  } else {

/**
 | Only the left- and right- pointers and the name field (using the
 | so called 'struct hack') are initialized
**/

    pNode = malloc(sizeof(symRec) + strlen(name));
    if (pNode) {
      pNode->left = pNode->right = 0;
      strcpy(pNode->name, name);
      if (pNew) *pNew = 1;
      *p = pNode;
    } else {
      fputs("No heap memory available\n", stderr);
      exit(EXIT_FAILURE);
    }
  }
  return pNode;
}

int yyerror(
  char *pc
){
  printf("  Error: %s\n", pc);
  return 0;
}

static void hex(void) {
  unsigned long dec = *last;

  printf("  %lu decimal = 0x%lX hex\n", dec, dec);
}

static void oct(void) {
  unsigned long dec = *last;

  printf("  %lu decimal = 0%lo octal\n", dec, dec);
}

static void printTree(void) {
  printf("  Current symbol table:\n");
  pN(root);
}

static void pN(
  const symRec *pNode
) {
  if (pNode->left) pN(pNode->left);

  switch (pNode->type) {
    case SYMBOL:
      printf("  - %s = %.10g\n", pNode->name, pNode->value.var);
      break;
    default:
      break;
  }

  if (pNode->right) pN(pNode->right);
}

static double myTrunc(
  double val
) {
  int i = val;

  return i;
}

static double frac(
  double val
) {
  int i = val;

  return fabs(val - i);
}

static double vround(
  double val
) {
  int i = floor(val + 0.5);

  return i;
}

static double myAsin(
  double rhs
) {
  if (rhs > 1.0   ||   rhs < -1.0) {
    puts("  Invalid argument");
    return 0;
  } else {
    return asin(rhs);
  }
}

static double myAcos(
  double rhs
) {
  if (rhs > 1.0   ||   rhs < -1.0) {
    puts("  Invalid argument");
    return 0;
  } else {
    return acos(rhs);
  }
}

static double myLog(
  double rhs
) {
  if (rhs <= 0.0) {
    puts("  Invalid argument");
    return 0;
  } else {
    return log(rhs);
  }
}

static double myLog10(
  double rhs
) {
  if (rhs <= 0.0) {
    puts("  Invalid argument");
    return 0;
  } else {
    return log10(rhs);
  }
}

static double mySqrt(
  double rhs
) {
  if (rhs < 0.0) {
    puts("  Invalid argument");
    return 0;
  } else {
    return sqrt(rhs);
  }
}

static double myAbs(
  double rhs
) {
  return rhs < 0 ? -rhs : rhs;
}
